"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const merklingSession_1 = __importDefault(require("./merklingSession"));
const symbols_1 = require("./symbols");
/**
 * The core IPLD access client.
 */
class Merkling {
    constructor(options) {
        this._ipfs = options.ipfs;
    }
    /**
     * Create a new merkling session to act as an
     * intermediary with the IPFS node.
     *
     * @returns {MerklingSession} a new merkling session
     */
    createSession() {
        return new merklingSession_1.default({ ipfs: this._ipfs });
    }
    /**
     * Run an action in the form of a callback over a new session
     * that is disposed of afterwards.
     *
     * @param sessionAction an action func that operates on the new session
     * @returns a promise that completes after the session action
     */
    withSession(sessionAction) {
        return __awaiter(this, void 0, void 0, function* () {
            const session = this.createSession();
            yield sessionAction(session);
        });
    }
    /**
     * Retrieve the CID for given proxy. Will returned
     * undefined if the passed object is not a Merkling proxy
     * and will return null if the proxy is dirty.
     *
     * @param proxy the object to get the CID for
     * @return the CID or undefined if the object will never
     * have a CID or null if it is currently dirty
     */
    static cid(proxy) {
        if (!Merkling.isProxy(proxy)) {
            return undefined;
        }
        if (!Merkling.isIpldNode(proxy)) {
            return undefined;
        }
        return proxy[symbols_1.getCidSymbol];
    }
    /**
     * Asynchronously force the retrieval of the state
     * stored in IPFS of the IPLD block the proxy represents.
     *
     * @param proxy An IPLD block proxy
     * @returns the given proxy now loaded and clean.
     */
    static resolve(proxy) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.isProxy(proxy)) {
                return proxy;
            }
            yield proxy[symbols_1.resolveSymbol];
            return proxy;
        });
    }
    /**
     * Retreive the stored state that the Merkling proxy
     * is representing. This will include references
     * to other Merkling proxies.
     *
     * @param proxy the proxy to inspect
     * @return the internal state of the IPLD block being represented
     */
    static inspect(proxy) {
        if (!this.isProxy(proxy)) {
            return proxy;
        }
        return proxy[symbols_1.getStateSymbol];
    }
    /**
     * Determine whether a given object is a tracked Merkling
     * proxy.
     *
     * @param potentialProxy
     * @returns the answer
     */
    static isProxy(potentialProxy) {
        if (!potentialProxy) {
            return false;
        }
        return !!potentialProxy[symbols_1.isProxySymbol];
    }
    /**
     * Determine whether a given object is a Merkling
     * proxy representing an IPLD block.
     *
     * @param potentialIpldNode
     * @returns the answer
     */
    static isIpldNode(potentialIpldNode) {
        if (!potentialIpldNode) {
            return false;
        }
        return !!potentialIpldNode[symbols_1.isIpldNodeSymbol];
    }
    /**
     * Determine if a proxy has unsaved changes. Will
     * return true if the passed object is not a proxy.
     *
     * @param proxy the object to test
     * @returns the answer
     */
    static isDirty(proxy) {
        if (!Merkling.isProxy(proxy)) {
            return true;
        }
        return !!proxy[symbols_1.isDirtySymbol];
    }
}
exports.Merkling = Merkling;
//# sourceMappingURL=merkling.js.map