"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const symbols_1 = require("./symbols");
const merkling_1 = require("./merkling");
var MerklingLifecycleState;
(function (MerklingLifecycleState) {
    MerklingLifecycleState["DIRTY"] = "DIRTY";
    MerklingLifecycleState["CLEAN"] = "CLEAN";
    MerklingLifecycleState["UNLOADED"] = "UNLOADED";
})(MerklingLifecycleState = exports.MerklingLifecycleState || (exports.MerklingLifecycleState = {}));
var MerklingProxyType;
(function (MerklingProxyType) {
    MerklingProxyType["IPLD"] = "IPLD";
    MerklingProxyType["INTERNAL"] = "INTERNAL";
})(MerklingProxyType = exports.MerklingProxyType || (exports.MerklingProxyType = {}));
class MerklingProxyRef {
    constructor(ref) {
        this.internalId = ref.internalId;
        this.type = ref.type;
        this.path = ref.path;
    }
    isRef() {
        return true;
    }
    toString() {
        if (this.path.length > 0) {
            return `${this.internalId}/${this.path.join('/')}`;
        }
        else {
            return this.internalId.toString();
        }
    }
}
exports.MerklingProxyRef = MerklingProxyRef;
const lookupRecordAndState = (target) => {
    const ipldRecord = target.session._ipldNodeEntries.get(target.ref.internalId);
    if (!ipldRecord) {
        return {
            record: undefined,
            state: undefined
        };
    }
    let state = ipldRecord.state;
    for (const key of target.ref.path) {
        state = state[key];
    }
    return {
        record: ipldRecord,
        state: state
    };
};
exports.merklingProxyHandler = {
    has: function (target, key) {
        const { state } = lookupRecordAndState(target);
        if (!state) {
            return false;
        }
        return key in state;
    },
    ownKeys(target) {
        const { state } = lookupRecordAndState(target);
        if (!state) {
            return [];
        }
        return Reflect.ownKeys(state);
    },
    getOwnPropertyDescriptor(target, key) {
        const { state } = lookupRecordAndState(target);
        if (!state) {
            return undefined;
        }
        return Object.getOwnPropertyDescriptor(state, key);
    },
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    get(target, key) {
        if (key === symbols_1.isProxySymbol) {
            return true;
        }
        if (key === symbols_1.isIpldNodeSymbol) {
            return target.ref.type === MerklingProxyType.IPLD;
        }
        if (key === symbols_1.getRefSymbol) {
            return target.ref;
        }
        if (key === symbols_1.resolveSymbol) {
            return target.session._resolveRef(target.ref);
        }
        const { record, state } = lookupRecordAndState(target);
        if (!record) {
            return undefined;
        }
        if (key === symbols_1.isDirtySymbol) {
            return record.lifecycleState === MerklingLifecycleState.DIRTY;
        }
        if (key === symbols_1.getStateSymbol) {
            return state;
        }
        if (key === symbols_1.getCidSymbol) {
            return record.cid;
        }
        if (key === symbols_1.getRecordSymbol) {
            return target;
        }
        // eslint-disable-next-line
        const value = state[key];
        if (!value || typeof value !== 'object') {
            return value;
        }
        let proxyId;
        if ('isRef' in value && value.isRef()) {
            proxyId = value;
            target.session._internalGraph.link(target.ref.internalId, proxyId.internalId);
        }
        else {
            proxyId = new MerklingProxyRef({
                internalId: target.ref.internalId,
                type: MerklingProxyType.INTERNAL,
                path: [key]
            });
        }
        if (target.session._proxies.has(proxyId.toString())) {
            return target.session._proxies.get(proxyId.toString());
        }
        const internalProxy = new Proxy({
            ref: proxyId,
            session: target.session
        }, exports.merklingProxyHandler);
        target.session._proxies.set(proxyId.toString(), internalProxy);
        return internalProxy;
    },
    set(target, key, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    value) {
        const { record, state } = lookupRecordAndState(target);
        if (!record || !state) {
            return false;
        }
        if (key === symbols_1.setCidSymbol) {
            if (record.type !== MerklingProxyType.IPLD) {
                throw new Error('Cannot set CID on internal proxy');
            }
            record.cid = value;
            record.lifecycleState = MerklingLifecycleState.CLEAN;
            return true;
        }
        else {
            target.session._markRecordAndAncestorsAsDirty(record.internalId);
            if (merkling_1.Merkling.isProxy(value)) {
                if (!merkling_1.Merkling.isIpldNode(value)) {
                    throw new Error('Setting sub-nodes not supported');
                }
                // eslint-disable-next-line
                const ref = value[symbols_1.getRefSymbol];
                target.session._internalGraph.link(target.ref.internalId, ref.internalId);
                return Reflect.set(state, key, ref);
            }
            else if (value != null && typeof value === 'object') {
                const { state: internalisedValue, childInternalIds } = target.session._internaliseState(value);
                for (const childInternalId of childInternalIds) {
                    target.session._internalGraph.link(record.internalId, childInternalId);
                }
                return Reflect.set(state, key, internalisedValue);
            }
            else {
                return Reflect.set(state, key, value);
            }
        }
    }
};
//# sourceMappingURL=merklingProxyHandler.js.map