"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const cids_1 = __importDefault(require("cids"));
const merklingProxyHandler_1 = require("./merklingProxyHandler");
const ipfsWrapper_1 = __importDefault(require("./ipfsWrapper"));
const serialiser_1 = __importDefault(require("./serialiser"));
const internalGraph_1 = __importDefault(require("./internalGraph"));
class MerklingSession {
    constructor({ ipfs }) {
        this._ipfs = new ipfsWrapper_1.default(ipfs);
        this._ipldIdCounter = 0;
        this._ipldNodeEntries = new Map();
        this._proxies = new Map();
        this._internalGraph = new internalGraph_1.default();
        this._serialiser = new serialiser_1.default((id) => {
            const record = this._ipldNodeEntries.get(id);
            if (!record) {
                throw new Error('Ref to unknown IPLD node - ' + id);
            }
            const cid = record.cid;
            if (!cid) {
                throw new Error('Attempted to serialize unsaved IPLD node');
            }
            return cid;
        });
    }
    /**
     * Create a new dirty Merkling proxy that can be manipulated
     * before later being persisted.
     *
     * @param objState a js object of the state for the IPLD block
     * @returns a proxy representing the given state as an IPLD block
     */
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    create(objState) {
        if (objState === null || typeof objState !== 'object') {
            return objState;
        }
        const { state, childInternalIds } = this._internaliseState(objState);
        const record = {
            internalId: ++this._ipldIdCounter,
            type: merklingProxyHandler_1.MerklingProxyType.IPLD,
            lifecycleState: merklingProxyHandler_1.MerklingLifecycleState.DIRTY,
            cid: null,
            state: state
        };
        const proxyId = new merklingProxyHandler_1.MerklingProxyRef({
            internalId: record.internalId,
            type: record.type,
            path: []
        });
        const proxy = new Proxy({
            ref: proxyId,
            session: this
        }, merklingProxyHandler_1.merklingProxyHandler);
        this._addIpldNodeEntry(record);
        for (const childInternalId of childInternalIds) {
            this._internalGraph.link(record.internalId, childInternalId);
        }
        // eslint-disable-next-line
        return proxy;
    }
    /**
     * Get an IPLD block from IPFS and return
     * it as a tracked Merkling proxy.
     * @param hashOrCid a CID or base encoded version
     * @returns a clean proxy with the state loaded and accessible
     */
    // eslint-disable-next-line
    get(hashOrCid) {
        return __awaiter(this, void 0, void 0, function* () {
            const cid = this._toCid(hashOrCid);
            const state = yield this._readStateFromIpfs(cid);
            const record = {
                internalId: ++this._ipldIdCounter,
                type: merklingProxyHandler_1.MerklingProxyType.IPLD,
                lifecycleState: merklingProxyHandler_1.MerklingLifecycleState.CLEAN,
                cid: cid,
                state: state
            };
            const proxyId = new merklingProxyHandler_1.MerklingProxyRef({
                internalId: record.internalId,
                type: record.type,
                path: []
            });
            this._addIpldNodeEntry(record);
            const proxy = new Proxy({
                ref: proxyId,
                session: this
            }, merklingProxyHandler_1.merklingProxyHandler);
            return proxy;
        });
    }
    load(hash) {
        const proxyId = this._loadCidIntoSession(hash);
        const proxy = new Proxy({
            ref: proxyId,
            session: this
        }, merklingProxyHandler_1.merklingProxyHandler);
        return proxy;
    }
    /**
     * Save all proxies that are currently marked as dirty,
     * and any proxies that depend on them.
     */
    save() {
        return __awaiter(this, void 0, void 0, function* () {
            for (const ipldNodeId of this._internalGraph.topologicalSort()) {
                const entry = this._ipldNodeEntries.get(ipldNodeId);
                if (!entry) {
                    throw new Error(`Failure of topological sort - no node with id ${ipldNodeId}`);
                }
                yield this._recursiveSaveIpldNode(entry);
            }
        });
    }
    /**
     * @private Given a merkling ref, retreive its state
     * from IPFS and update the internal record.
     *
     * @param ref a proxy ref to resolve
     * @returns a promise complete with the retrieval
     */
    _resolveRef(ref) {
        return __awaiter(this, void 0, void 0, function* () {
            const record = this._ipldNodeEntries.get(ref.internalId);
            if (!record) {
                return;
            }
            if (!record.cid) {
                throw new Error('Attempt to resolve ref without CID');
            }
            const state = yield this._readStateFromIpfs(record.cid);
            record.state = state;
            record.lifecycleState = merklingProxyHandler_1.MerklingLifecycleState.CLEAN;
        });
    }
    /**
     * @private Given an internal record mark it as dirty and
     * all loaded ancestors in the graph.
     *
     * @param internalId
     */
    _markRecordAndAncestorsAsDirty(internalId) {
        const recordAndAncestorsIds = this._internalGraph.ancestorsOf(internalId);
        let currentId = recordAndAncestorsIds.pop();
        if (!currentId) {
            throw Error('Internal graph returned no nodes in the ancestor lookup');
        }
        while (currentId !== undefined) {
            let currentRecord = this._ipldNodeEntries.get(currentId);
            if (!currentRecord) {
                throw new Error('Internal graph returned an internal id of a non-existant IPLD block');
            }
            currentRecord.lifecycleState = merklingProxyHandler_1.MerklingLifecycleState.DIRTY;
            currentRecord.cid = null;
            currentId = recordAndAncestorsIds.pop();
        }
    }
    // eslint-disable-next-line
    _internaliseState(obj) {
        const childIds = new Set();
        const internalLinker = (internalId) => {
            childIds.add(internalId);
        };
        const internalisedState = this._serialiser.internalise(obj, internalLinker);
        return {
            state: internalisedState,
            childInternalIds: Array.from(childIds)
        };
    }
    _loadCidIntoSession(hash) {
        const record = {
            internalId: ++this._ipldIdCounter,
            type: merklingProxyHandler_1.MerklingProxyType.IPLD,
            lifecycleState: merklingProxyHandler_1.MerklingLifecycleState.UNLOADED,
            cid: this._toCid(hash),
            state: undefined
        };
        const proxyId = new merklingProxyHandler_1.MerklingProxyRef({
            internalId: record.internalId,
            type: record.type,
            path: []
        });
        this._addIpldNodeEntry(record);
        return proxyId;
    }
    _addIpldNodeEntry(record) {
        this._ipldNodeEntries.set(record.internalId, record);
        this._internalGraph.add(record.internalId);
    }
    // eslint-disable-next-line
    _recursiveSaveIpldNode(record) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!record || record.type !== merklingProxyHandler_1.MerklingProxyType.IPLD) {
                throw new Error('Cannot recursively save a non-ipld node');
            }
            if (record.lifecycleState !== merklingProxyHandler_1.MerklingLifecycleState.DIRTY) {
                return;
            }
            const cid = yield this._saveRecordToIpfs(record);
            record.cid = cid;
            record.lifecycleState = merklingProxyHandler_1.MerklingLifecycleState.CLEAN;
        });
    }
    _saveRecordToIpfs(record) {
        return __awaiter(this, void 0, void 0, function* () {
            const serializedState = this._serialiser.serialise(record.state);
            return this._ipfs.put(serializedState);
        });
    }
    // eslint-disable-next-line
    _readStateFromIpfs(cid) {
        return __awaiter(this, void 0, void 0, function* () {
            const ipldNode = yield this._ipfs.get(cid.toBaseEncodedString());
            const substitutedState = this._serialiser.deserialise(ipldNode.value, (cid) => this._loadCidIntoSession(cid));
            return substitutedState;
        });
    }
    _toCid(hash) {
        if (typeof hash === 'string') {
            return new cids_1.default(hash);
        }
        else {
            return hash;
        }
    }
}
exports.default = MerklingSession;
//# sourceMappingURL=merklingSession.js.map